/********************************************************************
 MakeCR: Make Carriage Return

 copyright 1992 by Gordon Haff.  All rights reserved.

 3205 Windsor Ridge Dr.
 Westboro MA 01581
 Internet: 72561.2637@compuserve.com
 Compuserve: 72561,2637
 Channel 1 BBS, Cambridge, MA  (617) 354-8873

 Inserts a <CR>/<LF> at record_length intervals into a file.  A common
 reason to do this would be to add CR delimiters to a fixed-record
 datafile to allow processing by programs that require such delimiters
 (such as AWK or derivatives such as GAWK (Gnu AWK))

 Syntax: MAKCR in_file record_length [> out_file]

 in_file is the input file (usually fixed record length data)
 record_length is the interval at which to insert carriage returns

 The output will usually be redirected to an output file.  Otherwise,
 output will go to stdout -- usually the console.

 This program may be freely distributed so long as no charge is made for 
 its distribution except for customary and reasonable distribution
 charges.

 Written using Borland C++ 3.0 (ANSI C compiler mode) for MS-DOS systems

 Exit codes:
	0	normal exit
	2	incorrect number of arguments
	3	cannot open input file
	4	record_length appears to be illegal value
	5	could not allocate memory or 0 record_length

 ************************************************************************/

#include <stdio.h>
#include <dos.h>


FILE *in;

int main(int argc, char *argv[])	{


	char *buffer;
	int record_length;

	fputs("MakeCR 1.0.  Copyright 1992 by Gordon Haff. All Rights Reserved.\n\n",stderr);

	if (argc != 3) {
		fputs("Format: MAKCR in_file record_length [> out_file]\n\n",stderr);
		fputs("Inserts <CR>/<LF> combination into in_file every record_length\n",stderr);
		fputs("Use standard DOS redirection to output to a file, otherwise\n",stderr);
		fputs("stdout (unusually the console) will be used.\n\n",stderr);
		exit(2);
		}

	record_length = atoi(argv[2]);
	if (record_length <= 0)	{
		fputs("record_length appears to be illegal value\n\n",stderr);
		exit(4);
		}

	if (!(in=fopen(argv[1],"rb"))) {
		fputs("Cannot open specified input file\n\n",stderr);
		exit(3);
		}

	buffer = (char *)malloc(record_length+1);

	if (!buffer)	{
		fputs("Could not allocate memory or 0 record_length specified\n\n",stderr);
		exit(5);
		}

	while (!feof(in))	{

		fread(buffer,record_length,1,in);
		buffer[record_length] = 0;
		fputs(buffer,stdout);
		fputs("\n",stdout);

		}

	fclose(in);
	return 0;


}